''' Handles DSO table and observing list management, including alt-az/transit
    computations.
'''

import json
import time
import math
import numpy as np
from datetime import datetime, timedelta
from collections import Counter
from loguru import logger

from kivy.app import App
from kivy.clock import Clock
from kivy.core.window import Window
from kivy.properties import NumericProperty
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label
from kivy.metrics import dp
from kivymd.uix.slider import MDSlider

from jocular.component import Component
from jocular.settingsmanager import JSettings
from jocular.RA_and_Dec import RA, Dec
from jocular.table import Table, CButton, TableLabel
from jocular.calcs import local_sidereal_time, sun_altitude
from jocular.utils import toast


def ToHM(x):
    if math.isnan(x):
        return ''
    h = int(x)
    m = (x - h) * 60
    return f'{h:2.0f}h{m:02.0f}'.strip()


def fmt_diam(d):
    if d < 1:
        # ds = 60
        return f'{d * 60:4.1f}"'
    elif d < 10:
        return f"{d:4.2f}'"
    elif d < 100:
        return f"{d:4.1f}'"
    else:
        dg = int(d / 60)
        return f"{dg:.0f}\u00b0{(d - dg) / 60.0:2.0f}'"


def quadrant(x):
    # terrestrial direction
    qmap = ['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW']
    quad = int((x + 22.5) // 45)
    if quad == 8:
        quad = 0
    return qmap[quad]



class ObservingList(Component, JSettings):

    tab_name = 'Observatory'

    latitude = NumericProperty(42)
    longitude = NumericProperty(-2)

    configurables = [
        ('latitude', {
            'double_slider_float': (-90, 90),
            'fmt': '{:.2f}\u00b0',
            }),
        ('longitude', {
            'double_slider_float': (-180, 180),
            'fmt': '{:.2f}\u00b0',
            'help': 'Positive for points E of Greenwich, negative W'
            })
    ]


    def __init__(self):
        super().__init__()
        self.app = App.get_running_app()

        # time zone offset
        ts = time.time()
        self.utc_offset = (
            datetime.fromtimestamp(ts) - datetime.utcfromtimestamp(ts)
        ).total_seconds() / 3600.0

        Clock.schedule_once(self.load, 0)


    def load(self, dt=None):

        # v0.6 go back to treating self.objects as a view
        self.objects = Component.get('Catalogues').get_basic_dsos()

        # return a COPY of the basic dsos to prevent unwanted side effects??
        # self.objects = {k: v.copy() 
        #     for k, v in Component.get('Catalogues').get_basic_dsos().items()}

        # load observing list (maps names to date added)
        try:
            with open(self.app.get_path('observing_list.json'), 'r') as f:
                observing_list = json.load(f)
        except:
            observing_list = {}
        logger.info(f'{len(observing_list)} on observing list')

        # load observing notes
        try:
            with open(self.app.get_path('observing_notes.json'), 'r') as f:
                observing_notes = json.load(f)
        except:
            observing_notes = {}
        logger.info(f'{len(observing_notes)} observing notes')

        # load previous observations if necc and count previous
        try:
            obs = Component.get('Observations').get_observations()
            previous = dict(Counter([v['Name'].lower() for v in obs.values() if 'Name' in v]))
        except Exception as e:
            logger.warning(f'problem loading previous DSOs ({e})')
            previous = {}
        logger.info(f'{len(previous)} unique previous observations found')

        # augment/combine DSO info
        for v in self.objects.values():
            name = v['Name']
            v['Obs'] = previous.get(name.lower(), 0)
            v['Added'] = observing_list.get(name, '')
            v['List'] = 'Y' if name in observing_list else ''
            v['Notes'] = observing_notes.get(name, '')
            v['Other'] = v.get('Other', '')

        try:
            self.compute_transits()
        except Exception as e:
            logger.exception(f'problem computing transits {e}')


    def on_close(self):
        self.save_observing_list()
        self.save_notes()


    def save_observing_list(self, *args):
        try:
            ''' bug? if self.objects is changed (with new objects) then
                it won't have Added field. Temporary fix but needs migration
                to Catalogues
            '''
            ol = {}
            for v in self.objects.values():
                if v.get('Added', ''):
                    ol[v['Name']] = v['Added']
            with open(self.app.get_path('observing_list.json'), 'w') as f:
                json.dump(ol, f, indent=1)
        except Exception as e:
            logger.exception(f'problem saving ({e})')
            toast('problem saving observing list')


    def save_notes(self, *args):
        try:
            ol = {}
            for v in self.objects.values():
                if v.get('Notes', '').strip():
                    ol[v['Name']] = v['Notes']
            with open(self.app.get_path('observing_notes.json'), 'w') as f:
                json.dump(ol, f, indent=1)
        except Exception as e:
            logger.exception(f'problem saving ({e})')
            toast('problem saving observing notes')


    def compute_altaz(self, current_hours_offset=0):

        t = datetime.now() + timedelta(seconds=3600 * current_hours_offset)
        rads = math.pi / 180

        if hasattr(self, 'last_time_changed'):
            tdiff = t - self.last_time_changed
            if tdiff.seconds < 5:
                return
        self.last_time_changed = t

        lst = local_sidereal_time(t, self.longitude)
        lat = self.latitude * rads
        ras = np.array([v['RA'] for v in self.objects.values()])
        decs = np.array([v['Dec'] for v in self.objects.values()]) * rads

        sinlat, coslat = np.sin(lat), np.cos(lat)
        H = (-self.utc_offset * 15 + lst - ras) * rads
        sinH, cosH = np.sin(H), np.cos(H)
        az = 180 + (np.arctan2(sinH, cosH * sinlat - np.tan(decs) * coslat) / rads)
        alt = np.arcsin(sinlat * np.sin(decs) + coslat * np.cos(decs) * cosH) / rads

        for v, _alt, _az in zip(self.objects.values(), alt, az):
            v['Az'] = math.nan if math.isnan(_az) else int(_az)
            v['Alt'] = math.nan if math.isnan(_alt) else int(_alt)
            v['Quadrant'] = '' if math.isnan(_az) else quadrant(_az)


    def compute_transits(self):
        # from ch 15 of Meeus

        ras = np.array([v['RA'] for v in self.objects.values()])
        decs = np.array([v['Dec'] for v in self.objects.values()])

        #  apparent sidereal time at Greenwich at 0hUT
        t = datetime.now()
        gst_0 = local_sidereal_time(datetime(t.year, t.month, t.day), 0)

        max_alts = 90 - self.latitude + decs
        max_alts[max_alts > 90] = 180 - max_alts[max_alts > 90]

        # transit time: slight diff from Meeus since longitude negative from W here
        transiting = np.abs(decs) < (90 - self.latitude)
        m0 = (ras - self.longitude - gst_0 + self.utc_offset * 15) / 15
        m0[m0 > 24] = m0[m0 > 24] - 24
        m0[m0 < 0] = m0[m0 < 0] + 24

        # 80 ms or so
        for v, _max_alt, _transits, _m in zip(self.objects.values(), max_alts, transiting, m0):
            v['MaxAlt'] = math.nan if math.isnan(_max_alt) else int(_max_alt)
            v['Transit'] = float(_m) if _transits else math.nan

        logger.info(f'computed transits for {len(ras)} objects')


    def build(self):

        if not hasattr(self, 'objects'):
            self.load()

        cols = {
            'Name': {'w': 180, 'align': 'left', 'sort': {'catalog': ''}, 'action': self.new_from_list},
            'OT': {'w': 45},
            'Con': {'w': 45},
            'RA': {
                'w': 80,
                'align': 'right', 
                'type': float, 
                'val_fn': lambda x: x * 15, 
                'display_fn': lambda x: str(RA(x))
                },
            'Dec': {
                'w': 85, 
                'align': 'right', 
                'type': float, 
                'display_fn': lambda x: str(Dec(x))},
            'Quadrant': {'w': 40, 'align': 'center', 'label': 'Loc'},
            'Az': {'w': 40, 'align': 'center', 'type': float},
            'Alt': {'w': 40, 'align': 'center', 'type': float},
            'MaxAlt': {'w': 40, 'align': 'center', 'type': float, 'label': 'Max'},
            'Transit': {'w': 60, 'align': 'right', 'type': float, 'display_fn': ToHM},
            'Mag': {'w': 50, 'align': 'right', 'type': float},
            'Diam': {'w': 50, 'align': 'right', 'type': float, 'display_fn': fmt_diam},
            'Obs': {'w': 40, 'type': int},
            'List': {'w': 35},
            'Added': {'w': 65, 'sort': {'DateFormat': '%d %b'}},
            'Usr': {'w': 35, 'align': 'center'},
            'Notes': {'w': 100, 'input': True},
            'Other': {'w': 1, 'align': 'left'},
        }

        # time control widget
        ctrl = BoxLayout(orientation='horizontal', size_hint=(1, 1))

        # observing list buttons
        ctrl.add_widget(
            CButton(
                text='add to list', width=dp(100), on_press=self.add_to_observing_list
            )
        )
        ctrl.add_widget(
            CButton(
                text='del from list',
                width=dp(100),
                on_press=self.remove_from_observing_list,
            )
        )

        ctrl.add_widget(
            CButton(
                text='del from cat',
                width=dp(100),
                on_press=self.remove_from_catalogue,
            )
        )

        ctrl.add_widget(Label(size_hint_x=None, width=dp(40)))

        self.sun_time = TableLabel(text='', markup=True, size_hint_x=None, width=dp(80))

        ctrl.add_widget(self.sun_time)
        self.slider = MDSlider(
            orientation='horizontal',
            min=0,
            max=24,
            value=0,
            size_hint_x=None,
            width=dp(250),
        )
        self.slider.bind(value=self.time_changed)
        self.time_field = TableLabel(
            text=datetime.now().strftime('%d %b %H:%M'), size_hint_x=None, width=dp(160)
        )
        ctrl.add_widget(self.slider)
        ctrl.add_widget(self.time_field)
        self.time_changed(self.time_field, 0)

        logger.info('built')

        return Table(
            size=Window.size,
            data=self.objects,
            name='DSOs',
            cols=cols,
            update_on_show=False,
            controls=ctrl,
            on_hide_method=self.app.table_hiding,
            initial_sort_column='Name'
        )


    def time_changed(self, widgy, value):
        # user moves time slider
        if hasattr(self, 'update_event'):
            self.update_event.cancel()
        t0 = datetime.now() + timedelta(seconds=3600 * value)
        self.time_field.text = t0.strftime('%d %b %H:%M')

        sun_alt = sun_altitude(t0, self.latitude, self.longitude)
        self.sun_time.text = f'sun: {sun_alt:3.0f}\u00b0'
        try:
            self.compute_altaz(current_hours_offset=self.slider.value)
        except Exception as e:
            logger.exception(f'problem computing altaz {e}')

        if hasattr(self, 'table'):
            # update display since this is fast, but update table (reapply filters) when slider stops
            self.table.update_display()
            self.update_event = Clock.schedule_once(self.table.update, 0.5)


    @logger.catch()
    def show(self, *args):
        '''Called from menu to browse DSOs; open on first use'''
        if not hasattr(self, 'table'):
            self.table = self.build()
        self.app.showing = 'observing list'

        # redraw on demand when required
        if self.table not in self.app.gui.children:
            self.app.gui.add_widget(self.table, index=0)

        self.table.show()
        self.time_changed(self.time_field, 0)


    def new_from_list(self, row, *args):
        # User selects a row in the observing list table
        name = row.fields['Name'].text + '/' + row.fields['OT'].text
        self.table.hide()
        res = Component.get('Catalogues').lookup(name)
        if res is None:
            toast(f'Cannot find {name} in database')
        else:
            Component.get('DSO').new_DSO_name(res)


    def add_to_observing_list(self, *args):
        dn = datetime.now().strftime('%d %b')
        for s in self.table.selected:
            self.objects[s]['Added'] = dn
            self.objects[s]['List'] = 'Y'
        logger.info(f'added {len(self.table.selected)} objects')
        self.update_list()


    def remove_from_observing_list(self, *args):
        for s in self.table.selected:
            self.objects[s]['Added'] = ''
            self.objects[s]['List'] = ''
        logger.info(f'removed {len(self.table.selected)} objects')
        self.update_list()


    def remove_from_catalogue(self, *args):
        ''' Remove current DSO from user objects list
        '''
        catas = Component.get('Catalogues')
        for name in self.table.selected:
            obj = self.objects[name]
            # only delete if a user defined object
            if obj['Usr'] == 'Y':
                catas.delete_user_object(name)
        self.update_list()


    def new_observation(self):
        OT = Component.get('Metadata').get('OT', '')
        Name = Component.get('Metadata').get('Name', None)
        logger.info(f'{Name}/{OT}')
        #  update observed count if DSO is known
        if Name is not None:
            name = f'{Name}/{OT}'
            if name in self.objects:
                self.objects[name]['Obs'] = self.objects[name].get('Obs', 0) + 1
            if hasattr(self, 'table'):
                self.table.update()


    def update_list(self):
        self.save_observing_list()
        self.table.update()
        self.table.deselect_all()




""" Multi-frame bad pixel map
"""

import numpy as np
from scipy.ndimage import convolve
from loguru import logger

from kivy.app import App
from kivy.properties import BooleanProperty, NumericProperty

from jocular.component import Component
from jocular.settingsmanager import JSettings


class BadPixelMap(Component, JSettings):

    apply_BPM = BooleanProperty(True)
    sigmas = NumericProperty(5)
    bpm_frames = NumericProperty(3)

    tab_name = "Bad pixel map"
    configurables = [
        (
            "apply_BPM",
            {
                "name": "remove hot pixels?",
                "switch": "",
                "help": "Switching this off can help diagnose tracking issues",
            },
        ),
        (
            'bpm_frames', {
                'name': 'successive subs used to create BPM', 'float': (1, 10, 1),
                'help': 'Only treat as a bad pixel if it occurs in this many subs in succession',
                'fmt': '{:.0f} subs'
            },
        ),
        (
            "sigmas",
            {
                "name": "outlier rejection threshold",
                "float": (1, 10, 1),
                "help": "reject any pixel if this many sigmas from mean of 8-neighbourhood (factory: 5)",
                "fmt": "{:.0f} sigmas",
            },
        ),
    ]


    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()


    def on_new_object(self, *args):
        self.bpm = None
        self.bplist = []
        self.frame_count = 0


    def process_bpm(self, sub):
        ''' Called by Stacker
        '''

        if not self.apply_BPM:
            return

        im = sub.get_image()
        badpix = self.find_hot_pixels(im)

        # only update map if it is a light sub
        if sub.sub_type == 'light':
            self.update_bpm(badpix)
            if self.bpm is not None:
                logger.debug(f'{len(badpix)} pix, {len(self.bpm)} in map')
        self.do_bpm(im, self.bpm)


    def find_hot_pixels(self, im):
        """Return hot pixel candidates.
        Look for non-edge pixels whose intensity is significantly greater
        than their neighbours. Approach is conservative in order to
        find all hot pixels, since for EAA the only adverse effect is to
        replace a few non-hot pixels by their median. Returns a set of (row, col)
        coordinates.

        10ms for Lodestar, 50ms for ASI 290MM
        """

        # set min to zero
        im_norm = im - np.min(im)

        # divide im by local sum in 3x3 region
        im2 = im_norm / convolve(im_norm, np.ones((3, 3)), mode="constant")

        # Define hot pix as more than 'sigmas' SD from mean
        hp_cands = im2 > np.mean(im2) + self.sigmas * np.std(im2)

        # set boundaries to zero
        hp_cands[0, :] = 0
        hp_cands[-1, :] = 0
        hp_cands[:, 0] = 0
        hp_cands[:, -1] = 0

        # coordinates of hot pixels as 2 x N array
        hps = np.where(hp_cands > 0)

        return set(zip(hps[0], hps[1]))
        # return {(r, c) for r, c in zip(hps[0], hps[1])}


    def do_bpm(self, im, bpm=None):
        # Replace each pixel in bad pixel map by median of neighbours
        if bpm is None:
            bpm = self.bpm
        if bpm is not None:
            for r, c in bpm:
                im[r, c] = np.median(im[r-1:r+2, c-1:c+2].ravel())
        return im
 

    def compute_bpm(self):
        # Use intersection of bad pixels from previous N frames to compute bad pixel map
        if self.bplist:
            self.bpm = self.bplist[0]
            for bpl in self.bplist[1:]:
                self.bpm = self.bpm.intersection(bpl)


    def update_bpm(self, bpm):
        # Add new BPM set to BPM, recomputing BPM
        if len(self.bplist) == self.bpm_frames:
            self.bplist[self.frame_count % self.bpm_frames] = bpm
        else:
            self.bplist.append(bpm)
        self.frame_count += 1
        self.compute_bpm()
        

    def remove_hot_pixels(self, im):
        ''' Compute BPM and remove hot pixels in one operation, 
            without updating BPM. Used by calibrator
        '''
        hps = self.find_hot_pixels(im)
        im = im.copy()
        for r, c in hps:
            im[r, c] = np.median(im[r-1:r+2, c-1:c+2].ravel())
        return im





''' Home to star registration algorithms; currently
    two but only align_ransac is robust enough for the moment

    This is not part of the current Jocular release
'''

import warnings
import numpy as np
import math
from skimage.measure import ransac
from scipy.spatial.distance import cdist
from skimage.transform import EuclideanTransform, warp, estimate_transform


class AlignerException(Exception):
    pass


def align_ransac(im, keystars, centroids, min_stars=5, min_inliers=4, warp_model=None):
    ''' given a new image with centroids extracted, attempt to align
        against centroids represented by keystars
        return warped image
        raises AlignerException
    '''
    
    # we used to do this but it makes things worse...
    # if warp_model is not None:
    #     keystars = matrix_transform(keystars, warp_model.params)

    nstars = min(len(keystars), len(centroids))

    # find closest matching star to each transformed keystar
    matched_stars = np.zeros((nstars, 2))
    for i, (x1, y1) in enumerate(keystars):
        if i < nstars:
            matched_stars[i, :] = \
                centroids[np.argmin([(x1 - x2) ** 2 + (y1 - y2) ** 2 for x2, y2 in centroids])]

    # do we have enough matched stars?
    if len(matched_stars) < min_stars:
        raise AlignerException('not enough matched stars')
                
    # apply RANSAC to find which matching pairs best fitting Euclidean model
    # can throw a warning in cases where no inliers (bug surely) which we ignore
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        warp_model, inliers = ransac(
            (np.array(keystars[:nstars]), matched_stars),
            EuclideanTransform, 4, .5, max_trials=100)

        # if inliers is None:
        #     print('inliers is None')
        # elif sum(inliers) < min_inliers:
        #     print('sum inliers {:} less than min inliers {:}'.format(sum(inliers), min_inliers))

        # managed to align, so return warped image
        if (inliers is not None) and (sum(inliers) >= min_inliers):
            return warp(im, warp_model, order=3, preserve_range=True), warp_model

    raise AlignerException('not enough inliers after RANSAC')



def align_martin(im, keystars, centroids, min_stars=10):

    # assumes keystars and centroids are ordered by magnitude


    matches, _ = fastmatch(
        ref=keystars,
        im=centroids,
        match_threshold=5,  # pixels
        target_matches=25
    )

    if matches is None:
        raise AlignerException('no alignments found')
       
    src = keystars[[j for (i, j) in matches], :]
    dst = centroids[[i for (i, j) in matches], :]

    if len(src) < min_stars:
        raise AlignerException('not enough stars to align ({:} but need {:})'.format(len(src), min_stars))

    warp_model = estimate_transform('euclidean', src, dst)
    return warp(im, warp_model, order=3, preserve_range=True), warp_model
    

def fastmatch(
    ref=None,            # N x 2 array of centroids from keystars image
    im=None,             # N x 2 array of centroids from comparison image
    depth=5,             # points ordered by decreasing mag, so how far down do we go
    match_threshold=1 ,  # proximity in same units as centroids to claim a match
    target_matches=20,   # desired number of matches required before returning solution
    max_rot=90,           # degrees
    max_scale=100,        # percent
    max_translation=50   # in pixels
):
    ''' matches 2D point clouds allowing for rotatiom, translation and scaling
        limits on each of these can be set; if None, don't apply any limits
    '''
    max_matches = 0
    best_matches = None
    kk = match_threshold ** 2
    n_im, n_ref = len(im), len(ref)
    cnt = 0
    if max_scale is not None:
        min_scale = (100 - max_scale) / 100
        max_scale = (100 + max_scale) / 100

    for i1 in range(0, n_im):
        # normalise image based on star index i1
        im1 = im - im[i1, :]
        for i2 in [i for i in range(i1 + 1, min(n_im, i1 + depth)) if i != i1]:
            # rotate and scale so i2 is at (1, 0)
            cos, sin = im1[i2, 0], -im1[i2, 1]
            rot1 = np.degrees(math.atan2(sin, cos))
            d = cos ** 2 + sin ** 2
            im2 = np.dot(im1, np.array([[cos, -sin], [sin, cos]]).T) / d
            # match threshold takes into account scaling by d
            mt = kk / d  # we match squared dist
            min_x, min_y = np.min(im2, axis=0) - mt
            max_x, max_y = np.max(im2, axis=0) + mt
            for r1 in range(0, n_ref):
                ref1 = ref - ref[r1, :]
                for r2 in [
                    r for r in range(r1 + 1, min(n_ref, r1 + depth)) if r != r1
                ]:
                    cos, sin = ref1[r2, 0], -ref1[r2, 1]
                    rot2 = np.degrees(math.atan2(sin, cos))
                    # is rotation within limits?
                    if max_rot is None or np.abs(rot1 - rot2) < max_rot:
                        d2 = cos ** 2 + sin ** 2
                        scaling = (d/d2)**.5
                        # is scaling within limits?
                        if max_scale is None or scaling > min_scale and scaling < max_scale:
                            ref2 = np.dot(ref1, np.array([[cos, -sin], [sin, cos]]).T) / d2
                            mind_x, mind_y = np.min(ref2, axis=0)
                            maxd_x, maxd_y = np.max(ref2, axis=0)
                            # don't check if anything outside range
                            if (
                                max_x < maxd_x
                                and max_y < maxd_y
                                and min_x > mind_x
                                and min_y > mind_y
                            ):
                                cnt += 1
                                dists = cdist(im2, ref2, metric='sqeuclidean')
                                matches = [
                                    (i, j)
                                    for i, j in enumerate(np.argmin(dists, axis=1))
                                    if dists[i, j] < mt
                                ]
                                n_matches = len(matches)                            
                                if n_matches >= target_matches:
                                    return matches, cnt
                                if n_matches > max_matches:
                                    max_matches = n_matches
                                    best_matches = matches
        # return matches
    return best_matches, cnt





''' Manages all individual components, including imports/instantiation,
    in a uniform way.
'''

import json
import importlib
from collections import OrderedDict
from loguru import logger
from kivy.app import App
from kivy.event import EventDispatcher
from kivy.properties import StringProperty, ListProperty

class Component(EventDispatcher):

    components = OrderedDict()
    infoline = StringProperty('')   # component signals state for status panel
    save_settings = ListProperty([])
    changed = StringProperty('')


    @classmethod
    def get(cls, name):
        # Returns component whose class is identified by name; loads if not found
        if name in cls.components:
            return cls.components[name]
        try:
            module = importlib.import_module('jocular.' + name.lower())
            class_ = getattr(module, name)
            inst = class_()
            cls.register(inst)
            logger.info(f'{name} imported')
            return inst
        except Exception as e:
            logger.exception(f'cannot import module {name} ({e})')
            return None


    @classmethod
    def is_loaded(cls, name):
        return name in cls.components


    @classmethod
    def register(cls, obj):
        name = obj.__class__.__name__
        if name not in cls.components:
            cls.components[name] = obj
            App.get_running_app().gui.initialise_component(name)
        else:
            logger.error(f'name clash for {name}')


    @classmethod
    def initialise_new_object(cls):
        logger.info('initialising for new object')
        for v in cls.components.values():
            v.changed = ''
            v.on_new_object()


    @classmethod
    def initialise_previous_object(cls):
        # ensure metadata is first and stacker the last to be initialised
        logger.info('initialising for previous object')
        comps = ['Metadata'] + list(cls.components.keys() - {'Stacker', 'Metadata'}) + ['Stacker']
        for c in comps:
            cls.components[c].changed = ''
            cls.components[c].on_previous_object()


    @classmethod
    def changes(cls):
        ''' return a dictionary of changes since last save
        '''
        return {c: v.changed for c, v in cls.components.items() if v.changed != ''}


    @classmethod
    def save_object(cls):
        for v in cls.components.values():
            v.on_save_object()
            v.changed = ''


    @classmethod
    def close(cls):
        logger.debug('closedown')
        tosave = {}
        for c in cls.components.values():
            for k in c.save_settings:
                tosave[k] = getattr(c, k)
        with open(App.get_running_app().get_path('gui_settings.json'), 'w') as f:
            json.dump(tosave, f, indent=1)
        for c in cls.components.keys():
            cls.components[c].on_close()


    @classmethod
    def bind_status(cls):
        for name, inst in cls.components.items():
            cls.get('Status').bind_status(name, inst)


    @classmethod
    def check_for_change(cls):
        ''' Only changed if at least one component reports a change
            and the stack is non-empty
        '''
        pass


    @classmethod
    def any_changes(cls):
        nonempty = not cls.get('Stacker').is_empty()
        return (cls.changes() != {}) and nonempty


    def redraw(self, *args):
        pass


    def on_new_object(self):
        pass


    def on_previous_object(self, *args):
        # treat as new object unless a component overrides this
        self.on_new_object(*args)


    def on_save_object(self, *args):
        pass


    def on_close(self, *args):
        pass


    def on_changed(self, *args):
        ''' Called when any of the component 'changed' is altered
            Check if any changes from any component and communicate to GUI
        '''
        Component.check_for_change()


    def info(self, message=None):
        self.infoline = message



''' Jocular component that manages platesolving. Makes use of
    available platesolvers in platesolvers.py
'''


from loguru import logger

from kivy.properties import NumericProperty

from jocular.utils import toast
from jocular.component import Component
from jocular.settingsmanager import JSettings
from jocular.processing.platesolvers import platesolve, PlatesolverException


class PlateSolver(Component, JSettings):

    match_arcsec = NumericProperty(15)
    focal_length = NumericProperty(800)
    pixel_height = NumericProperty(8.4)
    n_stars_in_image = NumericProperty(30)
    min_sigma = NumericProperty(3)
    max_sigma = NumericProperty(5)
    star_thresh = NumericProperty(.001)
    min_matches = NumericProperty(10)
    target_matches = NumericProperty(10)
    binning = NumericProperty(1)

    tab_name = 'Plate-solving'

    configurables = [
        ('focal_length', {
            'name': 'focal length', 'float': (50, 2400, 10),
            'fmt': '{:.0f} mm',
            'help': 'Focal length of scope (including reducers)'
            }),
        ('pixel_height', {
            'name': 'pixel height', 
            'double_slider_float': (1, 16),
            'fmt': '{:.2f} um',
            'help': 'Pixel height of sensor; will be obtained from FITs where possible)'
            }),
        ('binning', {
            'name': 'binning', 'float': (1, 4, 1),
            'fmt': '{:.0f}',
            'help': 'If binning, specify amount here; will be obtained from FITs where possible (1=no binning)'
            }),
        ('n_stars_in_image', {
            'name': 'number of stars to extract', 'float': (5, 50, 1),
            'fmt': '{:.0f} stars',
            'help': 'Ideal number of stars to extract from image (factory: 30)'
            }),
        ('min_matches', {
            'name': 'minimum number of stars to match', 'float': (5, 30, 1),
            'fmt': '{:.0f} stars',
            'help': 'Must match at least this many stars in stellar database (factory: 10)'
            }),
        ('target_matches', {
            'name': 'target number of stars to match', 'float': (5, 30, 1),
            'fmt': '{:.0f} stars',
            'help': 'stop when we reach this many matches; set to same as min_matches to return first match (factory: 25)'
            }),
        ('match_arcsec', {
            'name': 'star proximity for matching', 'float': (5, 25, 1),
            'fmt': '{:.0f} arcsec',
            'help': 'How close reference and image stars must be to match, in arcsec (factory: 15)'
            })
        ]

    def on_new_object(self):
        self.info('')
        self.fov = None
        self.fwhm = None


    def describe(self):
        return {'fov': self.fov, 'fwhm': self.fwhm}


    def solve(self):
        ''' Called when user clicks 'solve'. Try to solve currently displayed image.
        '''

        star_db = Component.get('Catalogues').get_star_db()

        if star_db is None:
            toast('Cannot solve: no platesolving database available')
            return

        ra0, dec0 = Component.get('DSO').current_object_coordinates()
        if ra0 is None:
            toast('Cannot solve: unknown DSO')
            self.info('unknown DSO')
            return

        # get current image
        im = Component.get('Stacker').get_current_displayed_image()
        if im is None:
            toast('Cannot solve: no current image')
            self.info('no current image')
            return

        pixel_height = Component.get('Stacker').get_pixel_height()
        # if we don't know pixel height, use user-supplied values
        if pixel_height is None:
            pixel_height = self.binning * self.pixel_height

        dpp = (pixel_height * 1e-6 / self.focal_length) * (206265 / 3.6)

        try:
            soln = platesolve(
                im=im, 
                ra0=ra0, 
                dec0=dec0, 
                star_db=star_db,
                degrees_per_pixel=dpp,
                verbose=True,
                nstars=self.n_stars_in_image, 
                star_method='photutils', 
                min_matches=self.min_matches,
                match_arcsec=self.match_arcsec,
                target_matches=self.target_matches
                ) 

        except PlatesolverException as e:
            logger.warning(e)
            toast(str(e))
            self.info(str(e))
            return

        desc = f"{soln['fov_w']:5.3f} x {soln['fov_h']:5.3f}, {soln['north']:.0f}\u00b0, RA: {soln['ra']} Dec: {soln['dec']}"

        logger.info(desc)

        fwhm = soln['fwhm'] * dpp * 3600.
        self.fwhm = fwhm
        self.fov = f"{soln['fov_w']:.2f}\u00b0 x {soln['fov_h']:.2f}\u00b0"

        self.info('fwhm: {:.1f}" | fov: {:.2f}\u00b0 x {:.2f}\u00b0'.format(
            fwhm,
            soln['fov_w'], 
            soln['fov_h']))

        toast(desc, duration=3)

        Component.get('Annotator').annotate(soln)





import os


def connect_to_ASCOM_device(device_type=None, driver=None):
	''' Try to connect to ASCOM device; if driver is None, bring up chooser.
		Return dictionary containing some or all of driver, status, device and exception
	'''

	if os.name != 'nt':
		return {
			'status': 'ASCOM only works on Windows',
			'connected': False}

	try:
		import win32com.client
	except Exception as e:
		return {
			'status': 'Cannot import win30com.client; is ASCOM installed?',
			'connected': False,
			'exception': e
			}

	# try to connect to driver supplied
	if driver is not None:
		try:
			device = win32com.client.Dispatch(driver)
			device.Connected = True
			return {
				'device': device,
				'status': f'Connected to {device.Name}',
				'connected': True
				}
		except:
			pass # we'll give user a chance to choose

	# need a device type for chooser
	if device_type is None:
		return {
			'status': 'You must specify a device type',
			'connected': False
			}

	# chooser
	try:
		chooser = win32com.client.Dispatch("ASCOM.Utilities.Chooser")
		chooser.DeviceType = device_type
		driver = chooser.Choose(None)
	except Exception as e:
		return {
			'status': 'Unable to choose driver',
			'exception': e,
			'connected': False
			}

	# connect
	try:
		device = win32com.client.Dispatch(driver)
		device.Connected = True
		return {
			'device': device, 
			'driver': driver, 
			'status': f'Connected to {device.Name}',
			'connected': True
			}
	except Exception as e:
		return {
			'status': f'Unable to connect to {driver}',
			'exception': e,
			'connected': False
			}




""" Various astro calcs mainly based on Meuss. 
"""

import math
import time
import numpy as np
from datetime import datetime


def julian_date(when):
    # from Meuss p 61; 'when' is a datetime object

    y = when.year
    m = when.month
    d = when.day + when.hour / 24 + when.minute / (24 * 60) + when.second / (24 * 3600)

    if m < 3:
        y -= 1
        m += 12

    a = int(y / 100)

    if y >= 1582 and m >= 10:
        # Gregorian
        a = int(y / 100)
        b = 2 - a + int(a / 4)
    else:
        # Julian
        b = 0

    jd = int(365.25 * (y + 4716)) + int(30.6001 * (m + 1)) + d + b - 1524.5
    return jd


def to_range(x, d):
    # reduce x to range 0-d by adding or subtracting multiples of d
    if x < 0:
        return x - int((x / d) - 1) * d
    return x - int((x / d)) * d


def local_sidereal_time(when, longitude):
    # direct method of Meuss p87

    # when must be in UT
    jd = julian_date(when)
    t = (jd - 2451545.0) / 36525.0
    mst = (
        280.46061837
        + 360.98564736629 * (jd - 2451545.0)
        + 0.000387933 * t**2
        - t**3 / 38710000
    )

    # convert to 0-360
    mst = to_range(mst, 360)

    # convert from Greenwich to local
    lst = mst + longitude

    return lst


def sun_altitude(when, latitude, longitude):
    # Meuss p163+

    jd = julian_date(when)
    rads = math.pi / 180.0

    t = (jd - 2451545.0) / 36525.0
    L0 = 280.46646 + 36000.76983 * t + 0.0003032 * t * t
    L0 = to_range(L0, 360)
    M = 357.52911 + 35999.05029 * t - 0.0001537 * t * t
    # e = 0.016708634 - 0.000042037 * t - 0.0000001267 * t * t
    C = (
        (1.914602 - 0.004817 * t - 0.000014 * t * t) * np.sin(M * rads)
        + (0.019993 - 0.000101 * t) * np.sin(2 * M * rads)
        + 0.000289 * np.sin(3 * M * rads)
    )
    long_sun = L0 + C
    # v = M + C
    # R = (1.000001018 * (1 - e * e)) / (1 + e * np.cos(v * rads))
    sigma = 125.04 - 1934.136 * t
    lam = long_sun - 0.00569 - 0.00478 * np.sin(sigma * rads)
    ep = (
        23
        + (26 / 60)
        + (21.448 / 3600)
        - (46.815 * t + 0.00059 * t**2 - 0.001813 * t**3) / 3600
    )
    ep_corr = ep + 0.00256 * np.cos(sigma * rads)
    ra = (
        np.arctan2(np.cos(ep_corr * rads) * np.sin(lam * rads), np.cos(lam * rads))
        / rads
    )
    ra = to_range(ra, 360)
    dec = np.arcsin(np.sin(ep_corr * rads) * np.sin(lam * rads)) / rads

    # now convert to locale

    ts = time.time()
    utc_offset = (
        datetime.fromtimestamp(ts) - datetime.utcfromtimestamp(ts)
    ).total_seconds() / 3600.0
    lst = local_sidereal_time(when, longitude)
    lat = latitude * rads
    H = (-utc_offset * 15 + lst - ra) * rads
    alt = (
        np.arcsin(
            np.sin(lat) * np.sin(dec * rads)
            + np.cos(lat) * np.cos(dec * rads) * np.cos(H)
        )
        / rads
    )

    return alt
